	-------------------------------------------------------
-- Created by Ben Cope of Imperial College (May 2005)
-- cbcr Update Block for Primary Colour Correction
-------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

LIBRARY UNISIM;
USE UNISIM.vcomponents.ALL;

LIBRARY work;
USE work.Constants.ALL;
USE work.ALL;

 ENTITY CB_cbcr_alter IS
 PORT(
  -- Input Signals
  clock : IN STD_LOGIC;
  reset: IN STD_LOGIC;
  d_rotate_vector : IN Vector18;
  negate_angle : IN STD_LOGIC;
  sat_shift : IN Vector18;
  cb_in : IN Vector18;
  cr_in : IN Vector18;

  -- Output Signals
  cb_out : OUT Vector18;
  cr_out : OUT Vector18);
END CB_cbcr_alter;

ARCHITECTURE inside OF CB_cbcr_alter IS

SIGNAL scaled_cb, scaled_cr, cb_sqr, cr_sqr, new_gain : Vector36;
SIGNAL cb_in_cordic, cr_in_cordic : Vector19;
SIGNAL cb_out_cordic, cr_out_cordic : Vector19;
SIGNAL cb_cr, cb_sqr_small, cr_sqr_small, rounding_cb, rounding_cr, scaled_cb_small, scaled_cr_small: Vector18;
SIGNAL new_gain_hold : New_sat_hold_Array;

SIGNAL new_sat_3, new_sat_4, new_sat_5, new_sat_6 : Vector20;
SIGNAL sat_minus_max_dist, sat_too_big, sat_too_big_buffered : Vector20;

SIGNAL new_sat_7, new_sat_8, new_gain_small, new_gain_round : Vector18;
SIGNAL add_one : STD_LOGIC;
SIGNAL cb_cr_extended, sat_shift_extended : Vector20;

-- Constants specific to Input Correction
CONSTANT cordic_correct: Vector18 := "001001101101110101"; -- 1/1.6457 with 2I.16F
CONSTANT one : Vector18 := "010000000000000000"; -- format 2I.16F
CONSTANT max_dist_sqrt : Vector20 := "00001100001111111110"; -- format 19 +1 2I.18F
CONSTANT max_dist_sqrt_minus : Vector20 := "11110011110000000001"; -- format 19 +1 2I.18F
BEGIN

-- Cordic Rotation
cb_in_cordic <= cb_in(17) & cb_in;
cr_in_cordic <= cr_in(17) & cr_in;

Cordic1: ENTITY Cordic_rotate
 PORT MAP (
  clock => clock,
  reset => reset,
  negate => negate_angle,
  d => d_rotate_vector,
  x_in => cb_in_cordic,
  y_in => cr_in_cordic,
  x_out => cb_out_cordic,
  y_out => cr_out_cordic);

-- Post Cordic scaling (including Sat Shift)
 M2: MULT18X18S
 PORT MAP (
  P => scaled_cb,    -- 36-bit multiplier output
  A => new_gain_hold(9),    -- Cordic Correction*(1 + SatShift) format 2I.16F
  B => cb_out_cordic(18 DOWNTO 1),    -- CB returned format 2I.16F
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );
 M3: MULT18X18S
 PORT MAP (
  P => scaled_cr,    -- 36-bit multiplier output
  A => new_gain_hold(9),    -- Cordic Correction + SatGain	format 2I.16F
  B => cr_out_cordic(18 DOWNTO 1),    -- CR returned (consider rounding)
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );

-- Range Checking Approximation
 M4: MULT18X18S
 PORT MAP (
  P => cb_sqr,    -- 36-bit multiplier output
  A => cb_in,    -- cb input
  B => cb_in,    -- cb input
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );
 M5: MULT18X18S
 PORT MAP (
  P => cr_sqr,    -- 36-bit multiplier output
  A => cr_in,    -- cr input
  B => cr_in,    -- cr input
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );

 -- Pre-multiply cordic_correct and sat_gain(shifted)
 M7: MULT18X18S
 PORT MAP (
  P => new_gain,    -- 36-bit multiplier output	(max 13)
  A => cordic_correct,    -- cordic correct - format 2I.16F (1.64...)
  B => new_sat_8,    -- (1 + SatShift) - format 2I.16F 
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );


-- Main Process
P1 : PROCESS
VARIABLE new_sat_rounding, rounding_sat : Vector18;
VARIABLE not_ns_rounding: Vector18;

BEGIN

 WAIT UNTIL Rising_Edge(clock);

 IF (reset = '1') THEN
  cb_out <= (OTHERS => '0');
  cr_out <= (OTHERS => '0');
 ELSE

  -- Range Checking

  -- Cycle 1 - Multiply cb*cb and cr*cr

  -- Cycle 2 - Add results:
  cb_sqr_small <= cb_sqr(32 DOWNTO 15);
  cr_sqr_small <= cr_sqr(32 DOWNTO 15);

  -- Cycle 3
  cb_cr <= unsigned(cb_sqr_small) + unsigned(cr_sqr_small);  -- format 17 -2 signed

  -- Cycle 4 - Add SatShift
  cb_cr_extended <= "000" & cb_cr(17 DOWNTO 1);	-- format 2I.18F 19 +1
  sat_shift_extended <= sat_shift(17) & sat_shift(17) & sat_shift(17 DOWNTO 0); -- format 2I.18F 19 +1

  -- Cycle 5
  new_sat_3 <= signed(cb_cr_extended) + signed(sat_shift_extended);	-- format 2I.18F 19 +1

  -- Cycle 4 - Start Calculating possible new satshift for being too big
  sat_minus_max_dist <= signed(new_sat_3)	+ signed(max_dist_sqrt_minus);
  new_sat_4 <= new_sat_3;

  -- Cycle 5 - Start Calculating possible new satshift for being too big
  sat_too_big <= signed(sat_shift) - signed(sat_minus_max_dist);
  new_sat_5 <= new_sat_4;

  -- Buffer all
  sat_too_big_buffered <= sat_too_big;
  new_sat_6 <= new_sat_5;

  -- Cycle 7 Check if too big or small
  IF (signed(new_sat_6) < signed(max_dist_sqrt_minus)) THEN
   new_sat_7 <= (OTHERS => '0');
	add_one <= '0';
  ELSIF (signed(new_sat_6) > signed(max_dist_sqrt)) THEN
	new_sat_7 <= sat_too_big_buffered(19 DOWNTO 2);
	add_one <= '1';
  ELSE
	new_sat_7 <= sat_shift;
	add_one <= '1';
  END IF;

  -- Cycle 8 
  IF (add_one = '1') THEN
   new_sat_8 <= signed(new_sat_7) + signed(one);	 -- format 2I.16F
  ELSE
	new_sat_8 <= new_sat_7;  -- format 2I.16F
  END IF;
  
  -- Cycle 8 Multiply (1 + SatShift) to cordic correct
  
  new_gain_small <= new_gain(33 DOWNTO 16);
  new_gain_round <= "00000000000000000" & new_gain(15);

  -- Cycle 9
  new_gain_hold(0) <= signed(new_gain_small) + signed(new_gain_round);
  
 -- Cycles 10-22
 FOR i IN 0 TO 8 LOOP
  new_gain_hold(i+1) <= new_gain_hold(i);
 END LOOP;

 -- Multiply new cbcr by sat*cordic_correct

  -- Output rounded versions of cb and cr
  rounding_cb <= "00000000000000000" & scaled_cb(14);
  rounding_cr <= "00000000000000000" & scaled_cr(14);
  scaled_cb_small <= scaled_cb(32 DOWNTO 15);
  scaled_cr_small <= scaled_cr(32 DOWNTO 15);

  cb_out <= signed(scaled_cb_small) + signed(rounding_cb);
  cr_out <= signed(scaled_cr_small) + signed(rounding_cr);

 END IF;

END PROCESS P1;

END inside;
