-------------------------------------------------------
-- Created by Ben Cope of Imperial College (April 2005)
-- Constants and types for Primary Colour Correction
-------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

PACKAGE Constants IS

---------------------------------------------------------------------------------
 -- TYPEs for data buses
---------------------------------------------------------------------------------
 
 -- RGB I/O Signals
 CONSTANT RGB_width : INTEGER := 8;	-- unsigned
 CONSTANT Int_part : INTEGER := 0;
 CONSTANT Frac_part : INTEGER := RGB_width - Int_part;
 
 SUBTYPE Vector3 IS STD_LOGIC_VECTOR(2 DOWNTO 0);
 SUBTYPE Vector4 IS STD_LOGIC_VECTOR(3 DOWNTO 0);
  
 SUBTYPE VectorRGB IS STD_LOGIC_VECTOR(RGB_width-1 DOWNTO 0);
 TYPE VectorRGB_1x3 IS ARRAY (0 TO 2) OF VectorRGB;
 TYPE Array_VectorRGB_1x3 IS ARRAY (0 TO 5) OF VectorRGB_1x3;
 
 TYPE VectorRGB_4 IS ARRAY (0 TO 15) OF VectorRGB;
 TYPE VectorRGB_3x4 IS ARRAY (0 TO 2) OF VectorRGB_4;

 SUBTYPE Vector9 IS STD_LOGIC_VECTOR(8 DOWNTO 0);
 TYPE Vector9_1x3 IS ARRAY (0 TO 2) OF Vector9;

 SUBTYPE Vector12 IS STD_LOGIC_VECTOR(11 DOWNTO 0);
 TYPE Vector12_1x3 IS ARRAY (0 TO 2) OF Vector12;
 TYPE Matrix12_3x3 IS ARRAY (0 TO 2) OF Vector12_1x3;

 SUBTYPE Vector15 IS STD_LOGIC_VECTOR(14 DOWNTO 0);
 SUBTYPE Vector16 IS STD_LOGIC_VECTOR(15 DOWNTO 0);

 SUBTYPE Vector18 IS STD_LOGIC_VECTOR(17 DOWNTO 0);
 TYPE Vector18_1x3 IS ARRAY (0 TO 2) OF Vector18;
 TYPE Matrix18_3x3 IS ARRAY (0 TO 2) OF Vector18_1x3; 

 SUBTYPE Vector19 IS STD_LOGIC_VECTOR(18 DOWNTO 0);
 TYPE Vector19_1x3 IS ARRAY (0 TO 2) OF Vector19;

 SUBTYPE Vector20 IS STD_LOGIC_VECTOR(19 DOWNTO 0);
 SUBTYPE Vector21 IS STD_LOGIC_VECTOR(20 DOWNTO 0);

 SUBTYPE Vector22 IS STD_LOGIC_VECTOR(21 DOWNTO 0);

 SUBTYPE Vector36 IS STD_LOGIC_VECTOR(35 DOWNTO 0);
 TYPE Vector36_1x3 IS ARRAY (0 TO 2) OF Vector36;
 TYPE Matrix36_3x3 IS ARRAY (0 TO 2) OF Vector36_1x3;

 TYPE Cordic_Array IS ARRAY(0 TO 18) OF Vector36;
 TYPE Lum_delay_Array IS ARRAY(0 TO 18) OF Vector18;
 TYPE New_sat_hold_Array IS ARRAY(0 TO 18) OF Vector18;
 TYPE New_sat_abs_Array IS ARRAY(0 TO 1) OF STD_LOGIC_VECTOR(1 DOWNTO 0);
 
 SUBTYPE Vector256 IS bit_vector(255 DOWNTO 0);
 TYPE gamma_init_array IS ARRAY(0 TO 7) OF Vector256;

 ---------------------------------------------------------------------
 -- Constants 
 -- Format: explained as #bits (less sign bit) then 
 -- location of decimal point wrt right of sign bit
 ---------------------------------------------------------------------

 -- RGB to Ycbcr matrix 
 CONSTANT rgb_ycbcr_mx : Matrix18_3x3 :=  
 (("001000001110010101",   -- Format: 17  0
   "010000001000001100",   -- Format: 17  0
   "000011001000101101"),  -- Format: 17  0
  ("111011010000111001",   -- Format: 17  0
   "110110101100000010",   -- Format: 17  0
   "001110000011000100"),  -- Format: 17  0
  ("001110000011000100",   -- Format: 17  0
   "110100001110010101",   -- Format: 17  0
   "101101110100101111")); -- Format: 17 -3
-- | _11 _12 _13 |  | 0.257   0.504  0.098 | 
-- | _21 _22 _23 |  |-0.148  -0.291  0.439 | 
-- | _31 _32 _33 |  | 0.439  -0.368 -0.071 |  
  
  CONSTANT rgb_ycbcr_vec : Vector18_1x3 :=		
  ("000010000000000000", "010000000000000000", "010000000000000000");
-- Vector of following (correct)
-- (16.0/256.0, 128.0/256.0, 128.0/256.0);
 
 
  CONSTANT ycbcr_rgb_vec : Vector18_1x3 :=
  ("111110000000000000", "110000000000000000", "110000000000000000"); -- 2 integer 6 frac 
-- Vector of following (correct)
-- (-16.0/256.0, -128.0/256.0, -128.0/256.0);

 -- Ycbcr to RGB Matrix
-- CONSTANT ycbcr_rgb_mx : Matrix18_3x3 :=
-- (("001001010011111101",   -- Format: 17 +2
--   "000000000000000000",   -- Format: 17 +2
--   "001100110001001001"),  -- Format: 17 +2
--  ("001001010011111101",   -- Format: 17 +2
--   "100110111110011101",   -- Format: 17 -1
--   "111001011111101111"),  -- Format: 17 +2
--  ("001001010011111101",   -- Format: 17 +2
--	"010000001001001101",   -- Format: 17 +2
--	"000000000000000000")); -- Format: 17 +2
 
 CONSTANT ycbcr_rgb_mx : Matrix18_3x3 := -- Buffered to same form as RGB (ie 7 before + 1 after)		 
 (("001001010011111101", "000000000000000000", "001100110001001001"),
  ("001001010011111101", "111100110111110011", "111001011111101111"),
  ("001001010011111101", "010000001001001101", "000000000000000000")); -- 3 integer 7 frac
 -- | _11 _12 _13 |  | 1.164  0.0    1.596 |
 -- | _21 _22 _23 |  | 1.164 -0.391 -0.813 |
 -- | _31 _32 _33 |  | 1.164  2.018  0.0   |
  
 -- Convinient Vectors: 
 CONSTANT RGB_zeros : VectorRGB := (OTHERS => '0');
 CONSTANT zerosRGB_1x3 : VectorRGB_1x3 :=
 ((OTHERS => '0'), (OTHERS => '0'), (OTHERS => '0'));
 CONSTANT zeros12_1x3 : Vector12_1x3 :=
 ((OTHERS => '0'), (OTHERS => '0'), (OTHERS => '0'));
 CONSTANT zeros18_1x3 : Vector18_1x3 :=
 ((OTHERS => '0'), (OTHERS => '0'), (OTHERS => '0'));
 CONSTANT zeros18 : Vector18 := (OTHERS => '0');
 CONSTANT ones12 : Vector12 := (OTHERS => '1');
 CONSTANT one_12 : Vector12 := "001000000000";
 CONSTANT RGB_ones : VectorRGB := (OTHERS => '1');  -- 2 integer 8 frac (may cause errors - check out) 
 
 -- Test-Bench Related
 CONSTANT clock_period : TIME := 10 ns;

 CONSTANT rgb_increment : VectorRGB_1x3 :=
 ("00000001", "00000001", "00000001");
 CONSTANT xyl_increment : Vector12_1x3 := 
 ("000000000001", "000000000001", "000000000001");

 -- Delay of each block
 CONSTANT rgb_xyl_latency: INTEGER := 6;
 CONSTANT input_corr_latency: INTEGER := 24;
 CONSTANT cordic_latency: INTEGER := 23;
 CONSTANT xyl_rgb_latency: INTEGER := 6;
 CONSTANT total_latency_ic: INTEGER := rgb_xyl_latency +  input_corr_latency + xyl_rgb_latency + 3;
 CONSTANT histogram_latency: INTEGER := 8;
 CONSTANT colour_bal_latency: INTEGER := cordic_latency + 3; 
 CONSTANT total_latency_cb: INTEGER := rgb_xyl_latency +  colour_bal_latency + xyl_rgb_latency;
 CONSTANT primary_ccr_latency: INTEGER := total_latency_ic + histogram_latency + total_latency_cb;
   
 TYPE Holding_RGB_Array IS ARRAY (0 TO rgb_xyl_latency + xyl_rgb_latency + cordic_latency + 3) OF VectorRGB_1x3;
 TYPE Holding_Lum_Array IS ARRAY (0 TO cordic_latency + xyl_rgb_latency + 3) OF Vector18;  

 CONSTANT gamma_init : gamma_init_array := 		-- gamma = 1 addresses right to left in columns
 (x"AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
  x"CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC",
  x"F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0F0",
  x"FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00FF00",
  x"FFFF0000FFFF0000FFFF0000FFFF0000FFFF0000FFFF0000FFFF0000FFFF0000",
  x"FFFFFFFF00000000FFFFFFFF00000000FFFFFFFF00000000FFFFFFFF00000000",
  x"FFFFFFFFFFFFFFFF0000000000000000FFFFFFFFFFFFFFFF0000000000000000",
  x"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000000000000000000000000000");  

  
END Constants;

