-------------------------------------------------------------
-- Created by Ben Cope of Imperial College (April 2005)
-- Histogram Equalisation Block for Primary Colour Correction
-------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

LIBRARY UNISIM;
USE UNISIM.vcomponents.ALL;

LIBRARY work;
USE work.Constants.ALL;

ENTITY Histogram_Equalisation IS
PORT(
  -- Input Signals
  clock : IN STD_LOGIC;
  ram_clock: IN STD_LOGIC;
  reset: IN STD_LOGIC; 
  RGB_in : IN VectorRGB_1x3;	-- format 7 -1
  black_level : IN Vector9;  -- format 8 0
  BW_DIFF : IN Vector18;	  -- format 17 +4
  equalise : IN Vector3;
  output_black_level : IN VectorRGB;
  output_white_level : IN VectorRGB;

  -- Update signals for gamma RAM
  addr_update : IN VectorRGB;
  write_enable : IN STD_LOGIC;
  gamma_update : IN VectorRGB_1x3;

  -- Output Signals
  RGB_out : OUT VectorRGB_1x3
);
END Histogram_Equalisation;

ARCHITECTURE inside OF Histogram_Equalisation IS

SIGNAL RGB_original : Array_VectorRGB_1x3;
SIGNAL RGB_0 : Vector18_1x3;
SIGNAL RGB_1, RGB_1_del : Vector36_1x3;
SIGNAL RGB_2, RGB_in_ext : Vector9_1x3;
SIGNAL RGB_4, RGB_4_del, RGB_5, RGB_5_1 : VectorRGB_1x3;
SIGNAL WE : Vector16;
SIGNAL RGB_3, RGB_null : VectorRGB_3x4;
--SIGNAL w_clock : STD_LOGIC;

CONSTANT one : STD_LOGIC_VECTOR(4 DOWNTO 0) := "00001";

BEGIN

G1: FOR i IN 0 TO 2 GENERATE	  -- Loop through RGB channels
 MULT18X18S_inst : MULT18X18S
  PORT MAP (
   P => RGB_1(i),    -- 36-bit multiplier output
   A => RGB_0(i),    -- RGB Channel
   B => BW_DIFF,    -- Constant Matrix
   C => clock,    -- Clock input
   CE => '1',  -- Clock enable input
   R => reset     -- Synchronous reset input
  );
END GENERATE;

 R1: FOR i IN 0 TO 2 GENERATE		  -- RAM Implementation
  R2: FOR j IN 7 DOWNTO 0 GENERATE
   R3: FOR k IN 0 TO 15 GENERATE
	  WE(k) <= write_enable WHEN (k = conv_integer(unsigned(addr_update(7 DOWNTO 4)))) ELSE '0';
	  RAM16X1D_1_inst : RAM16X1D_1
      generic map (
       INIT => gamma_init(j)((k+1)*16-1 DOWNTO k*16))
      port map (
       DPO => RGB_3(i)(k)(j), -- Port A 1-bit data output
       SPO => open,           -- Port B 1-bit data output
      A0 => addr_update(0),  -- Port A address[0] input bit
      A1 => addr_update(1),  -- Port A address[1] input bit
      A2 => addr_update(2),  -- Port A address[2] input bit
      A3 => addr_update(3),  -- Port A address[3] input bit
      D => gamma_update(i)(j),-- Port A 1-bit data input
      DPRA0 => RGB_2(i)(0),  -- Port B address[0] input bit
      DPRA1 => RGB_2(i)(1),  -- Port B address[1] input bit
      DPRA2 => RGB_2(i)(2),  -- Port B address[2] input bit
      DPRA3 => RGB_2(i)(3),  -- Port B address[3] input bit
      WCLK => ram_clock,       -- Port A write clock input
      WE => WE(k)            -- Port A write enable input
      );
	END GENERATE;
  END GENERATE;
 END GENERATE;

P1: PROCESS
VARIABLE rgb_less_bl : Vector9_1x3;

BEGIN

 WAIT UNTIL Rising_Edge(clock);

 IF (reset = '1') THEN

  RGB_out <= zerosRGB_1x3;
 
 ELSE



  FOR i IN 0 TO 2 LOOP
  -- Cycle 1
   RGB_in_ext(i) <= "0" & RGB_in(i);

  -- Cycle 2
	rgb_less_bl(i) := unsigned(RGB_in_ext(i)) - unsigned(black_level);
	RGB_0(i) <= "00000"& rgb_less_bl(i) & "0000";
  END LOOP;


  -- Cycle 3 - Multiply by BW_DIFF
  RGB_1_del <= RGB_1;
  
  -- Cycle 4 - Range Limit to [0, 1]
  FOR i IN 0 TO 2 LOOP
   IF (RGB_1_del(i)(35) = '1') THEN
	 RGB_2(i) <= (OTHERS => '0');
	ELSIF	(unsigned(RGB_1_del(i)(29 DOWNTO 25)) >= signed(one)) THEN
	 RGB_2(i)(7 DOWNTO 0) <= (OTHERS => '1');
	 RGB_2(i)(8) <= '0';
	ELSE
	 RGB_2(i) <= RGB_1_del(i)(25 DOWNTO 17);
	END IF;  
  END LOOP;
  
 --  Cycle 5 - Look-up RGB_2^gamma   
  RGB_4(0) <= RGB_3(0)(conv_integer(unsigned(RGB_2(0)(7 DOWNTO 4))));
  RGB_4(1) <= RGB_3(1)(conv_integer(unsigned(RGB_2(1)(7 DOWNTO 4))));
  RGB_4(2) <= RGB_3(2)(conv_integer(unsigned(RGB_2(2)(7 DOWNTO 4))));

  
  -- Cycle 6 - Range Limit
  FOR i IN 0 TO 2 LOOP
   IF (unsigned(RGB_4(i)) > unsigned(output_white_level)) THEN
    RGB_5_1(i) <= output_white_level;
   ELSE
	 RGB_5_1(i) <= RGB_4(i);
   END IF;
  END LOOP;

  RGB_4_del <= RGB_4;

  FOR i IN 0 TO 2 LOOP
   IF (unsigned(RGB_4_del(i)) < unsigned(output_black_level)) THEN
	 RGB_5(i) <= output_black_level;
   ELSE
	 RGB_5(i) <= RGB_5_1(i);
	END IF;
  END LOOP;
  
  -- Cycle 7 - Select Between original and new
  FOR i IN 0 TO 2 LOOP
   IF (equalise(i) = '1') THEN
	 RGB_out(i) <= RGB_5(i);
   ELSE
	 RGB_out(i) <= RGB_original(4)(i);
	END IF;
  END LOOP;

 END IF;
   
END PROCESS P1;

delayed_rgb: PROCESS

BEGIN
 WAIT UNTIL Rising_Edge(clock);
  FOR i IN 0 TO 4 LOOP
   RGB_original(i+1) <= RGB_original(i);
  END LOOP;
  RGB_original(0) <= RGB_in;
END PROCESS delayed_rgb;

END inside;