-------------------------------------------------------
-- Created by Ben Cope of Imperial College (May 2005)
-- Luminance Update Block for Primary Colour Correction
-------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

LIBRARY UNISIM;
USE UNISIM.vcomponents.ALL;

LIBRARY work;
USE work.Constants.ALL;

ENTITY IC_Lum_alter IS
 PORT(
  -- Input Signals
  clock : IN STD_LOGIC;
  reset: IN STD_LOGIC;
  lum_shift : IN Vector12;
  lum_gain : IN Vector15;
  lum_in : IN Vector18;

  -- Output Signals
  Lum_out : OUT Vector18);
END IC_Lum_alter;

ARCHITECTURE inside OF IC_Lum_alter IS
SIGNAL Lum_2 : Vector18;
SIGNAL Lum_1 : Vector19;
SIGNAL scaled_lum: Vector36;
SIGNAL Lum_extended : Vector19;
SIGNAL lum_shift_int : Vector19;
SIGNAL lum_gain_int : Vector18;
SIGNAL Lum_delay : Lum_delay_Array;
SIGNAL scaled_lum_comp, scaled_lum_comp_1 : Vector22;
SIGNAL scaled_lum_rounded, scaled_lum_small : Vector18;
SIGNAL rounding : Vector18;

CONSTANT lower_limit : Vector19 := "0000010000000000000"; -- 16/256 with 2I.17F
CONSTANT upper_limit : Vector19 := "0011101011000000000"; -- 235/256 with 2I.17F
CONSTANT lower_limit_ls : Vector22 := "0000000010000000000000"; -- 16/235 with 5I.17F
CONSTANT upper_limit_ls : Vector22 := "0000011101011000000000"; -- 235/256 with 5I.17F

BEGIN

-- Multiplier for Luminance Scaling
M1: MULT18X18S
 PORT MAP (
  P => scaled_lum,    -- 36-bit multiplier output
  A => Lum_2,    -- Luminance
  B => lum_gain_int,    -- lum_gain - multiplied by 2^-4
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );

 -- Main Process
P1 : PROCESS

BEGIN

 WAIT UNTIL Rising_Edge(clock);

 lum_shift_int <= lum_shift(11) & lum_shift & "000000";
 lum_gain_int <= "0" & lum_gain & "00";

 IF (reset = '1') THEN
  
  Lum_out <= (OTHERS => '0');

 ELSE

  -- LEVEL 1 PIPELINE
  
  -- Apply luminance gain
  Lum_extended <= Lum_in(17) & Lum_in;

  Lum_1 <= signed(Lum_extended) + signed(lum_shift_int);  

  -- LEVEL 2 PIPELINE

  -- Range Limit
  IF (signed(Lum_1) < signed(lower_limit)) THEN
    Lum_2 <= lower_limit(17 DOWNTO 0);
  ELSIF (signed(Lum_1) > signed(upper_limit)) THEN
    Lum_2 <= upper_limit(17 DOWNTO 0);
  ELSE
    Lum_2 <= Lum_1(17 DOWNTO 0);
  END IF;

  -- LEVEL 3 PIPELINE

  -- Perform Luminance Multiply

  -- LEVEL 4 Range Limit	 - may pipline extra later
  rounding <= "00000000000000000" & scaled_lum(12);
  scaled_lum_small <= scaled_lum(30 DOWNTO 13);
  scaled_lum_comp_1 <= scaled_lum(34 DOWNTO 13); 

  scaled_lum_comp <= scaled_lum_comp_1;
  scaled_lum_rounded <= signed(scaled_lum_small) + signed(rounding);

  -- LEVEL 5 Compare
  IF (signed(scaled_lum_comp) < signed(lower_limit)) THEN
    Lum_delay(0) <= lower_limit(17 DOWNTO 0);
  ELSIF (signed(scaled_lum_comp) > signed(upper_limit_ls)) THEN
    Lum_delay(0) <= upper_limit_ls(17 DOWNTO 0);
  ELSE
    Lum_delay(0) <= scaled_lum_rounded;
  END IF;	

  -- DELAY Lum by 16 cycles
  FOR i IN 0 TO 15 LOOP
   Lum_delay(i+1) <= Lum_delay(i);
  END LOOP;

  Lum_out <= Lum_delay(16);
 END IF;

END PROCESS P1;

END inside;

