	-------------------------------------------------------
-- Created by Ben Cope of Imperial College (May 2005)
-- cbcr Update Block for Primary Colour Correction
-------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

LIBRARY UNISIM;
USE UNISIM.vcomponents.ALL;

LIBRARY work;
USE work.Constants.ALL;
USE work.ALL;

 ENTITY IC_cbcr_alter IS
 PORT(
  -- Input Signals
  clock : IN STD_LOGIC;
  reset: IN STD_LOGIC;
  d_rotate_vector : IN Vector18;
  negate_angle : IN STD_LOGIC;
  sat_gain : IN Vector18;
  cb_in : IN Vector18;
  cr_in : IN Vector18;

  -- Output Signals
  cb_out : OUT Vector18;
  cr_out : OUT Vector18);
END IC_cbcr_alter;

ARCHITECTURE inside OF IC_cbcr_alter IS

SIGNAL scaled_cb, scaled_cr, cb_sqr, cr_sqr, new_sat, new_gain : Vector36;
SIGNAL cb_in_cordic, cr_in_cordic : Vector19;
SIGNAL cb_out_cordic, cr_out_cordic : Vector19;
SIGNAL cb_cr, new_sat_rounded, new_sat_neg, rounding_cb, rounding_cr, cb, cr: Vector18;
SIGNAL new_sat_small, rounding_4, cb_sqr_small, cr_sqr_small, new_gain_small, rounding_sat : Vector18;
SIGNAL new_sat_hold : New_sat_hold_Array;
SIGNAL new_sat_abs : New_sat_abs_Array;

-- Constants specific to Input Correction
--CONSTANT cordic_correct: Vector19 := "0010011011011101010"; -- 1/1.6457 with 2I.16F
CONSTANT cordic_correct: Vector18 := "001001101101110101"; -- 1/1.6457 with 2I.16F
CONSTANT one : Vector18 := "000000000000000001";

BEGIN

-- Cordic Rotation
cb_in_cordic <= cb_in(17) & cb_in;
cr_in_cordic <= cr_in(17) & cr_in;

Cordic1: ENTITY Cordic_rotate
 PORT MAP (
  clock => clock,
  reset => reset,
  negate => negate_angle,
  d => d_rotate_vector,
  x_in => cb_in_cordic,
  y_in => cr_in_cordic,
  x_out => cb_out_cordic,
  y_out => cr_out_cordic);

-- Post Cordic scaling (including Sat Gain)
 M2: MULT18X18S
 PORT MAP (
  P => scaled_cb,    -- 36-bit multiplier output
  A => new_sat_hold(15),    -- Cordic Correction + SatGain format 5I.13F
  B => cb_out_cordic(18 DOWNTO 1),    -- CB returned format 2I.16F
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );
 M3: MULT18X18S
 PORT MAP (
  P => scaled_cr,    -- 36-bit multiplier output
  A => new_sat_hold(15),    -- Cordic Correction + SatGain	format 5I.13F
  B => cr_out_cordic(18 DOWNTO 1),    -- CR returned (consider rounding)
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );

-- Range Checking Approximation
 M4: MULT18X18S
 PORT MAP (
  P => cb_sqr,    -- 36-bit multiplier output
  A => cb_in,    -- cb input
  B => cb_in,    -- cb input
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );
 M5: MULT18X18S
 PORT MAP (
  P => cr_sqr,    -- 36-bit multiplier output
  A => cr_in,    -- cr input
  B => cr_in,    -- cr input
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );
 M6: MULT18X18S
 PORT MAP (
  P => new_sat,    -- 36-bit multiplier output format 36 -2	(2^-7 down)
  A => cb_cr,    -- cb_sqr + cr_sqr	 format 18 -2
  B => sat_gain,    -- Saturation Gain	*2^-7
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );

 -- Pre-multiply cordic_correct and sat_gain(shifted)
 M7: MULT18X18S
 PORT MAP (
  P => new_gain,    -- 36-bit multiplier output	(max 13)
  A => cordic_correct,    -- cordic correct - format 2I.16F (1.64...)
  B => new_sat_hold(3),    -- Saturation Gain - format 5I.16F (max 8 min 0)
  C => clock,    -- Clock input
  CE => '1',  -- Clock enable input
  R => reset     -- Synchronous reset input
 );

-- Main Process
P1 : PROCESS
VARIABLE new_sat_rounding : Vector18;
VARIABLE not_ns_rounding: Vector18;

BEGIN

 WAIT UNTIL Rising_Edge(clock);

 IF (reset = '1') THEN
  cb_out <= (OTHERS => '0');
  cr_out <= (OTHERS => '0');
 ELSE

  -- Range Checking

  -- Cycle 1 - Multiply cb*cb and cr*cr

  -- Cycle 2 - Add results:
  cb_sqr_small <= cb_sqr(32 DOWNTO 15);
  cr_sqr_small <= cr_sqr(32 DOWNTO 15);

  -- Cycle 3
  cb_cr <= unsigned(cb_sqr_small) + unsigned(cr_sqr_small);  -- format 18 -2 signed

  -- Cycle 4 - Multiply	by Sat Gain

  -- Cycle 5 - Truncate:
  rounding_4 <= "00000000000000000" & new_sat(15);
  new_sat_small <= new_sat(33 DOWNTO 16);

  -- Cycle 6
  new_sat_rounded <= signed(new_sat_small) + signed(rounding_4);	-- only interested in integer part

  -- Cycle 7 - Find Negative of Sat
  not_ns_rounding := not(new_sat_rounded);
  new_sat_neg <= signed(not_ns_rounding) + signed(one);
  new_sat_hold(0) <= new_sat_rounded;

  -- Cycle 8 Check if negative and take abs
  IF (new_sat_hold(0)(17) = '1') THEN
   new_sat_abs(0) <= new_sat_neg(16 DOWNTO 15);
  ELSE
	new_sat_abs(0) <= new_sat_hold(0)(16 DOWNTO 15);
  END IF;
  new_sat_hold(1) <= sat_gain;

  -- Cycle 9 If abs(value) = 1 or more right shift sat by 2 else do nothing
  IF (new_sat_abs(0)(1) = '1') THEN
   new_sat_hold(2) <= new_sat_hold(1)(17) & new_sat_hold(1)(17) & new_sat_hold(1)(17 DOWNTO 2);
  ELSE
	new_sat_hold(2) <= new_sat_hold(1);
  END IF;
  new_sat_abs(1) <= new_sat_abs(0);

  -- Cycle 10 - If abs(value) = 0.5 or more right shift by 1 else do nothing
  IF (new_sat_abs(1)(0) = '1') THEN
   new_sat_hold(3) <= new_sat_hold(2)(17) & new_sat_hold(2)(17 DOWNTO 1);
  ELSE
	new_sat_hold(3) <= new_sat_hold(2);
  END IF;

  -- Cycle 11 - Multiply Sat Gain to Cordic Correct

  -- Cycle 12
  new_gain_small <= new_gain(33 DOWNTO 16);
  rounding_sat <= "00000000000000000" & new_gain(15);

  -- Cycle 13 - round result
  
 new_sat_hold(4) <= signed(new_gain_small) + signed(rounding_sat); -- 5I.13F

 -- Cycles 14-22
 FOR i IN 4 TO 14 LOOP
  new_sat_hold(i+1) <= new_sat_hold(i);
 END LOOP;

 -- Multiply new cbcr by sat*cordic_correct

  -- Output rounded versions of cb and cr
  cb <= scaled_cb(29 DOWNTO 12);
  cr <= scaled_cr(29 DOWNTO 12);
  rounding_cb <= "00000000000000000" & scaled_cb(11);
  rounding_cr <= "00000000000000000" & scaled_cr(11);

  cb_out <= signed(cb) + signed(rounding_cb);
  cr_out <= signed(cr) + signed(rounding_cr);

 END IF;

END PROCESS P1;

END inside;
