-------------------------------------------------------
-- Created by Ben Cope of Imperial College (April 2005)
-- RGB to XYL conversion for Primary Colour Correction
-------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.std_logic_arith.ALL;

LIBRARY UNISIM;
USE UNISIM.vcomponents.ALL;

LIBRARY work;
USE work.Constants.ALL;

ENTITY rgb_to_xyl IS
 PORT(
  -- Input Signals 
  clock : IN STD_LOGIC;
  reset : IN STD_LOGIC;
  RGB : IN VectorRGB_1x3;
  
  -- Output Signals
  XYL : OUT Vector18_1x3);
END rgb_to_xyl;
  
ARCHITECTURE inside OF rgb_to_xyl IS

CONSTANT lower_limit : Vector18 := "000010000000000000"; -- 16/256 with 1I.17F
CONSTANT upper_limit_y : Vector18 := "011101011000000000"; -- 235/256 with 1I.17F
CONSTANT upper_limit_cbcr : Vector18 := "011110000000000000"; -- 240/256 with 1I.17F
CONSTANT param1 : Vector18 := "010000000000000000"; -- 128/256 with 1I.17F

SIGNAL x: Matrix18_3x3;
SIGNAL rounding: Vector12_1x3;
SIGNAL result: Matrix36_3x3;
SIGNAL minterm0, minterm1, ycbcr, ycbcr_limited, ycbcr_nonrd: Vector18_1x3;
SIGNAL A: Vector18_1x3;

BEGIN

 -----------------------------------------------------------
 -- Concurrent Assignments - Buffer RGB signals with zeros
 -----------------------------------------------------------
 
 A(0)(17) <= '0';
 A(0)(16 DOWNTO 16-RGB_width+1) <= RGB(0);
 A(0)(16-RGB_width DOWNTO 0) <= (OTHERS => '0');
 A(1)(17) <= '0';
 A(1)(16 DOWNTO 16-RGB_width+1) <= RGB(1);
 A(1)(16-RGB_width DOWNTO 0) <= (OTHERS => '0');
 A(2)(17) <= '0';
 A(2)(16 DOWNTO 16-RGB_width+1) <= RGB(2);
 A(2)(16-RGB_width DOWNTO 0) <= (OTHERS => '0');	  

 ------------------------------------------------------------
 -- Multipliers for Matrix
 ------------------------------------------------------------
 G1: FOR j IN 0 TO 2 GENERATE	  -- Loop through RGB channels
  G2: FOR i IN 0 TO 2 GENERATE	-- Loop through Matrix
    MULT18X18S_inst : ENTITY Multiplier_model
    PORT MAP (
     P => result(i)(j),    -- 36-bit multiplier output
     A => A(j),    -- RGB Channel
     B => rgb_ycbcr_mx(i)(j),    -- Constant Matrix
     C => clock,    -- Clock input
     CE => '1',  -- Clock enable input
     R => reset     -- Synchronous reset input
    );
  END GENERATE;
 END GENERATE;


 Main_proc : PROCESS
  VARIABLE a,b,c: STD_LOGIC_VECTOR(0 TO 2);
   
  BEGIN

   WAIT UNTIL Rising_Edge(clock);
   
   IF (reset = '1') THEN

    XYL <= zeros18_1x3;
	 ycbcr_limited <= zeros18_1x3;
    ycbcr <= zeros18_1x3;
	 minterm0 <= zeros18_1x3;
	 minterm1 <= zeros18_1x3;

   ELSE

 -------------------------------------------------------------
 -- Perform Additions on Matrix Products - output named ycbcr
 -------------------------------------------------------------

	  -- Reduce to same # bits as input (no delay)
	  FOR i IN 0 TO 2 LOOP
	   FOR j IN 0 TO 2 LOOP
		 x(i)(j) <= result(i)(j)(34 DOWNTO 17);	
      END LOOP;
	  END LOOP;

	  x(2)(2) <= result(2)(2)(35) & result(2)(2)(35) & result(2)(2)(35 DOWNTO 20);

	  FOR i IN 0 TO 1 LOOP

		a(i) := result(i)(0)(16);
		b(i) := result(i)(1)(16);
		c(i) := result(i)(2)(16);
		rounding(i) <= "0000000000" & ((a(i) and b(i)) or (b(i) and c(i)) or (c(i) and a(i))) & 
		                               (a(i) xor b(i) xor c(i));
     END LOOP;

	  a(2) := result(2)(0)(16);
	  b(2) := result(2)(1)(16);
	  c(2) := result(2)(2)(19);
	  rounding(2) <= "0000000000" & ((a(2) and b(2)) or (b(2) and c(2)) or (c(2) and a(2))) & 
		                               (a(2) xor b(2) xor c(2));


	  -- PIPELINED ADDITION 
	  FOR i IN 0 TO 2 LOOP
	   minterm0(i) <= signed(x(i)(0)) + signed(x(i)(1));
		minterm1(i) <= signed(x(i)(2)) + signed(rgb_ycbcr_vec(i));
     END LOOP;

	  FOR i IN 0 TO 2 LOOP
	   ycbcr_nonrd(i) <= signed(minterm0(i)) + signed(minterm1(i));
     END LOOP;

	  FOR i IN 0 TO 2 LOOP
	   ycbcr(i) <= signed(ycbcr_nonrd(i)) + signed(rounding(i));
     END LOOP;

	
 -------------------------------------------------------------
 -- Range Limit Result
 ------------------------------------------------------------- 
		  
    -- y														  
    IF (unsigned(ycbcr(0)) < unsigned(lower_limit)) THEN
     ycbcr_limited(0) <= lower_limit;
    ELSIF (unsigned(ycbcr(0)) > unsigned(upper_limit_y)) THEN
     ycbcr_limited(0) <= upper_limit_y;
    ELSE	
	  ycbcr_limited(0) <= ycbcr(0);
    END IF;
    
    -- cb
    IF (signed(ycbcr(1)) < signed(lower_limit)) THEN
     ycbcr_limited(1) <= lower_limit;
    ELSIF (signed(ycbcr(1)) > signed(upper_limit_cbcr)) THEN
     ycbcr_limited(1) <= upper_limit_cbcr; 
	 ELSE
     ycbcr_limited(1) <= ycbcr(1);
    END IF;

    -- cr
    IF (signed(ycbcr(2)) < signed(lower_limit)) THEN
     ycbcr_limited(2) <= lower_limit;
    ELSIF (signed(ycbcr(2)) > signed(upper_limit_cbcr)) THEN
     ycbcr_limited(2) <= upper_limit_cbcr;
    ELSE
	  ycbcr_limited(2) <= ycbcr(2);
    END IF;    
 
    -- convert ycbcr to XYL space
    XYL(0) <= signed(ycbcr_limited(1)) - signed(param1);
    XYL(1) <= signed(ycbcr_limited(2)) - signed(param1);
    XYL(2) <= ycbcr_limited(0);
    
    END IF;
    
  END PROCESS Main_proc;
  
END inside;
  
  